from copy import deepcopy
import unittest

import torch
from torch import tensor
from torch.autograd import Variable
from torch.nn import Parameter
from torch.nn.modules import Linear, MSELoss
from torchvision import models

from rotational_update import RotationalLinear


if torch.cuda.is_available():
    device = 'cuda'
else:
    device = 'cpu'


class TestRotationalLinearNet(unittest.TestCase):
    def setUp(self) -> None:
        torch.manual_seed(0)
        self.vgg16 = models.vgg16().to(device)
        self.vgg16.classifier[-1] = Linear(4096, 10, bias=True).to(device)
        self.rotated_vgg16 = deepcopy(self.vgg16)

    # 同期式のレイヤーから、パラメータをそのままにして準同期式レイヤを作る
    def testParam(self):
        self.rotated_vgg16.classifier[0] = RotationalLinear(
            self.rotated_vgg16.classifier[0],
        ).to(device)
        self.rotated_vgg16.classifier[3] = RotationalLinear(
            self.rotated_vgg16.classifier[3],
        ).to(device)

        # 別オブジェクトだが同じパラメータを持つことを確認
        self.assertTrue(self.vgg16 is not self.rotated_vgg16)

        for layer1, layer2 in zip(self.vgg16.classifier, self.rotated_vgg16.classifier):
            self.assertTrue(isinstance(layer1, Linear) == isinstance(layer2, Linear))

            if isinstance(layer1, Linear) and isinstance(layer2, Linear):
                self.assertTrue((layer1.weight == layer2.weight).byte().all())
                self.assertTrue((layer1.bias == layer2.bias).byte().all())


class TestRotationalLinear(unittest.TestCase):
    def setUp(self) -> None:
        torch.manual_seed(0)
        base_layer = Linear(2, 4)
        base_layer.weight = Parameter(tensor([
            [1.7, 0.4, 1, 2.2],
            [1.8, -1, 0.9, -0.2]
        ], requires_grad=True).t())
        base_layer.bias = Parameter(tensor([0.0, 0.0, 0.0, 0.0], requires_grad=True))

        self.layer = RotationalLinear(base_layer).to(device)

    # def testGroup(self):
    #     self.assertEqual([0, 2, 4], self.layer.group_delim)

    def testForward(self):
        X = tensor([  # データ数3と仮定, shape=(3, 2)
            [1.2, 1.4],
            [-0.8, 0.2],
            [1.1, -1.2]
        ]).to(device)
        expected = tensor([
            [4.56, -0.92, 2.46, 2.36],
            [-1, -0.52, -0.62, -1.8],
            [-0.29, 1.64, 0.02, 2.66]
        ]).to(device)
        actual = self.layer.forward(X)
        self.assertTrue(torch.all(torch.abs(expected - actual) < 10e-6))

    def testBackward(self) -> None:
        X = Variable(tensor([  # データ数3と仮定, shape=(3, 2)
            [1.2, 1.4],
            [-0.8, 0.2],
            [1.1, -1.2]
        ], requires_grad=True).to(device))
        X.requires_grad_()

        expected_dx = tensor([
            [5.3, 1.5],
            [5.3, 1.5],
            [5.3, 1.5]
        ]).to(device)

        # -- neuron 0,1 get gradient --
        expected_dW = tensor([
            [1.5, 1.5, 0, 0],
            [0.4, 0.4, 0, 0]
        ]).t().to(device)
        expected_db = tensor([3.0, 3.0, 0, 0]).to(device)

        dout = self.layer.forward(X)
        dout.backward(torch.ones((3, 4)).to(device))

        self.assertTrue(
            torch.all(torch.abs(expected_dx - X.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_dW - self.layer.weight.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_db - self.layer.bias.grad) < 10e-6).item())

        self.layer.rotate()

        # clear gradients
        self.layer.weight.grad = None
        self.layer.bias.grad = None
        X.grad = None

        # -- neuron 2,3 get gradient --
        expected_dW = tensor([
            [0, 0, 1.5, 1.5],
            [0, 0, 0.4, 0.4]
        ]).t().to(device)
        expected_db = tensor([0, 0, 3.0, 3.0]).to(device)

        dout = self.layer.forward(X)
        dout.backward(torch.ones((3, 4)).to(device))

        self.assertTrue(
            torch.all(torch.abs(expected_dx - X.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_dW - self.layer.weight.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_db - self.layer.bias.grad) < 10e-6).item())

        self.layer.rotate()

        # clear gradients
        self.layer.weight.grad = None
        self.layer.bias.grad = None
        X.grad = None

        # -- neuron 0,1 get gradient again --
        expected_dW = tensor([
            [1.5, 1.5, 0, 0],
            [0.4, 0.4, 0, 0]
        ]).t().to(device)
        expected_db = tensor([3.0, 3.0, 0, 0]).to(device)

        dout = self.layer.forward(X)
        dout.backward(torch.ones((3, 4)).to(device))
        self.assertTrue(
            torch.all(torch.abs(expected_dx - X.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_dW - self.layer.weight.grad) < 10e-6).item())
        self.assertTrue(
            torch.all(torch.abs(expected_db - self.layer.bias.grad) < 10e-6).item())

    def testRotate(self):
        X = tensor([  # データ数3と仮定, shape=(3, 2)
                    [1.2, 1.4],
                    [-0.8, 0.2],
                    [1.1, -1.2]
        ], requires_grad=True).to(device)

        self.layer.forward(X)
        left = self.layer.learn_l
        right = self.layer.learn_r
        self.assertEqual(0, left)
        self.assertEqual(2, right)
        self.layer.rotate()

        self.layer.forward(X)
        left = self.layer.learn_l
        right = self.layer.learn_r
        self.assertEqual(2, left)
        self.assertEqual(4, right)
        self.layer.rotate()

        self.layer.forward(X)
        left = self.layer.learn_l
        right = self.layer.learn_r
        self.assertEqual(0, left)
        self.assertEqual(2, right)
        self.layer.rotate()


class TestRotationalLinearAtUnbalance(unittest.TestCase):
    def setUp(self) -> None:
        torch.manual_seed(0)
        self.layer = RotationalLinear(Linear(4000, 4000)).to(device)

    def testGroup(self):
        expected = [63 * i for i in range(64)] + [4000]
        self.assertEqual(expected, self.layer.group_partition)

    def testRotate(self):
        X = torch.ones((1, 4000)).to(device)

        for i in range(63):
            self.layer.forward(X)
            left = self.layer.learn_l
            right = self.layer.learn_r
            self.assertEqual(63 * i, left)
            self.assertEqual(63 * (i + 1), right)
            self.layer.rotate()

        self.layer.forward(X)
        left = self.layer.learn_l
        right = self.layer.learn_r
        self.assertEqual(3969, left)
        self.assertEqual(4000, right)
        self.layer.rotate()

        self.layer.forward(X)
        left = self.layer.learn_l
        right = self.layer.learn_r
        self.assertEqual(0, left)
        self.assertEqual(63, right)
        self.layer.rotate()


if __name__ == '__main__':
    unittest.main()



import torch
from torch.autograd import Function
from torch.nn.functional import linear
from torch import Tensor


class RotationalLinearFunction(Function):
    @staticmethod
    def forward(ctx, *args) -> Tensor:
        x, w, b, learn_left, learn_right = args
        learn_l = torch.as_tensor(learn_left).requires_grad_(False)
        learn_r = torch.as_tensor(learn_right).requires_grad_(False)

        ctx.save_for_backward(x, w, b, learn_l, learn_r)
        return linear(x, w, b)

    @staticmethod
    def backward(ctx, *args: Tensor) -> tuple:
        grad = args[0]
        x, w, b, learn_l, learn_r = ctx.saved_tensors
        w = w.t()
        learn_l, learn_r = int(learn_l), int(learn_r)

        # バイアスへの勾配は、0ベクトルを作って必要な要素だけ値を入れる
        # gradients for bias, make 0 vector and insert value into needed element
        if grad.is_cuda:
            d_b = torch.zeros(size=(grad.shape[1],), dtype=torch.float32, device='cuda')
        else:
            d_b = torch.zeros(size=(grad.shape[1],), dtype=torch.float32)
        d_b[learn_l:learn_r] = torch.sum(grad[:, learn_l:learn_r], dim=0)

        # 重みへの勾配は、0行列を作って必要な行だけ値を入れる
        # gradients for weights, make 0 matrix and insert value into needed column
        if grad.is_cuda:
            d_w = torch.zeros(size=(x.shape[1], grad.shape[1]), dtype=torch.float32, device='cuda')
        else:
            d_w = torch.zeros(size=(x.shape[1], grad.shape[1]), dtype=torch.float32)
        d_w[:, learn_l:learn_r] = torch.matmul(x.t(), grad[:, learn_l:learn_r])

        d_x = torch.matmul(grad, torch.t(w))

        return d_x, d_w.t(), d_b, None, None



from copy import deepcopy
from math import sqrt

from torch import Tensor
from torch.nn.modules import Linear

from rotational_update.layers.base import Rotatable
from rotational_update.layers.functions.static import RotationalLinearFunction


class RotationalLinear(Linear, Rotatable):
    """
    Rotatable Linear class
    """
    def __init__(self, linear: Linear):
        """
        Parameters
        ----------
        linear : torch.nn.modules.Linear
            Base Linear object. RotatationalLinear object takes over weights and biases from base object.
        """
        super().__init__(linear.in_features, linear.out_features, bias=True)
        # 重みを引き継ぎ
        # use current weights
        self.weight = deepcopy(linear.weight)
        self.bias = deepcopy(linear.bias)
        self.learn_l = None
        self.learn_r = None

        output_features = linear.out_features

        # group size -> sqrt(output_features)
        num = int(sqrt(output_features))
        group_list = [num for _ in range(num)]
        nueron_sum = sum(group_list)

        if nueron_sum < output_features:  # make one more group with extra neurons
            group_list.append(output_features - nueron_sum)

        # cumulative sum for slicing
        group_partition = [0] + group_list
        for i in range(1, len(group_partition)):
            group_partition[i] += group_partition[i - 1]

        group_i = 1

        self.group_partition = group_partition
        self.group_i = group_i

    def forward(self, input_tensor) -> Tensor:
        """
        Feed-forward method
        Almost same to normal Linear object.
        Save variables for learning group.

        Parameters
        ----------
        input_tensor : torch.Tensor

        Returns
        -------
        res : Output of feed-forwarding.
        """
        self.learn_l = self.group_partition[self.group_i-1]
        self.learn_r = self.group_partition[self.group_i]

        matmul = RotationalLinearFunction.apply
        res = matmul(input_tensor, self.weight, self.bias, self.learn_l, self.learn_r)

        return res

    def rotate(self):
        """
        Change learning group to next one.
        Call this at every minibatch spending.
        """
        self.group_i += 1
        if self.group_i == len(self.group_partition):
            self.group_i = 1  # 最初のグループへ  to initial group



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="rotational_update",
    packages=[
        "rotational_update",
        "rotational_update.layers",
        "rotational_update.layers.functions"
        ],

    version="0.0.17",
    license="MIT",

    install_requires=["torch>=1.3.0",],

    author="Tetsuya Hori",
    author_email="axis38akasira@gmail.com",

    url="https://github.com/ryhoh/Rotational_update",

    description="Rotational_update module for PyTorch",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='rotational rotational-update',

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
